/** transfer.c
 **
 ** Omnigo/Unix file transfer program.
 **
 ** Written by and Copyright 1996 Asher Hoskins, except for the sections
 ** from the GEOS SDK (marked).
 ** Modified by Marcus Groeber for use under OS/2.
 **
 ** The author retains copyright on this implementation. Permission for
 ** educational and non-profit use is granted to all. If you're planning to
 ** make money with this or any code derived from it, check with the author
 ** first.
 **
 ** See README file for more information.
 **
 ** File transfer routines.
 **/


#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef OS2
	#include <io.h>
#endif

#include "reply.h"

#define OG_ACK			0
#define OG_NAK			1
#define OG_NAK_QUIT		2
#define OG_SYNC			0xff
#define BLOCK_START		1
#define BLOCK_END		2
#define BLOCK_QUOTE		3

unsigned char readByte(int);
unsigned short readWord(int);
unsigned int readDWord(int);
void writeByte(int, unsigned char);
void writeWord(int, unsigned short);
void writeDWord(int, unsigned int);
unsigned short incCRC(unsigned short, char);
unsigned short calcCRC(char *, short, unsigned short);

/**
 ** This table is straight out of the GEOS-SDK.
 **/

short crcTable[] = {
	0x0000,0x1021,0x2042,0x3063,0x4084,0x50a5,0x60c6,0x70e7,0x8108,0x9129,
	0xa14a,0xb16b,0xc18c,0xd1ad,0xe1ce,0xf1ef,0x1231,0x0210,0x3273,0x2252,
	0x52b5,0x4294,0x72f7,0x62d6,0x9339,0x8318,0xb37b,0xa35a,0xd3bd,0xc39c,
	0xf3ff,0xe3de,0x2462,0x3443,0x0420,0x1401,0x64e6,0x74c7,0x44a4,0x5485,
	0xa56a,0xb54b,0x8528,0x9509,0xe5ee,0xf5cf,0xc5ac,0xd58d,0x3653,0x2672,
	0x1611,0x0630,0x76d7,0x66f6,0x5695,0x46b4,0xb75b,0xa77a,0x9719,0x8738,
	0xf7df,0xe7fe,0xd79d,0xc7bc,0x48c4,0x58e5,0x6886,0x78a7,0x0840,0x1861,
	0x2802,0x3823,0xc9cc,0xd9ed,0xe98e,0xf9af,0x8948,0x9969,0xa90a,0xb92b,
	0x5af5,0x4ad4,0x7ab7,0x6a96,0x1a71,0x0a50,0x3a33,0x2a12,0xdbfd,0xcbdc,
	0xfbbf,0xeb9e,0x9b79,0x8b58,0xbb3b,0xab1a,0x6ca6,0x7c87,0x4ce4,0x5cc5,
	0x2c22,0x3c03,0x0c60,0x1c41,0xedae,0xfd8f,0xcdec,0xddcd,0xad2a,0xbd0b,
	0x8d68,0x9d49,0x7e97,0x6eb6,0x5ed5,0x4ef4,0x3e13,0x2e32,0x1e51,0x0e70,
	0xff9f,0xefbe,0xdfdd,0xcffc,0xbf1b,0xaf3a,0x9f59,0x8f78,0x9188,0x81a9,
	0xb1ca,0xa1eb,0xd10c,0xc12d,0xf14e,0xe16f,0x1080,0x00a1,0x30c2,0x20e3,
	0x5004,0x4025,0x7046,0x6067,0x83b9,0x9398,0xa3fb,0xb3da,0xc33d,0xd31c,
	0xe37f,0xf35e,0x02b1,0x1290,0x22f3,0x32d2,0x4235,0x5214,0x6277,0x7256,
	0xb5ea,0xa5cb,0x95a8,0x8589,0xf56e,0xe54f,0xd52c,0xc50d,0x34e2,0x24c3,
	0x14a0,0x0481,0x7466,0x6447,0x5424,0x4405,0xa7db,0xb7fa,0x8799,0x97b8,
	0xe75f,0xf77e,0xc71d,0xd73c,0x26d3,0x36f2,0x0691,0x16b0,0x6657,0x7676,
	0x4615,0x5634,0xd94c,0xc96d,0xf90e,0xe92f,0x99c8,0x89e9,0xb98a,0xa9ab,
	0x5844,0x4865,0x7806,0x6827,0x18c0,0x08e1,0x3882,0x28a3,0xcb7d,0xdb5c,
	0xeb3f,0xfb1e,0x8bf9,0x9bd8,0xabbb,0xbb9a,0x4a75,0x5a54,0x6a37,0x7a16,
	0x0af1,0x1ad0,0x2ab3,0x3a92,0xfd2e,0xed0f,0xdd6c,0xcd4d,0xbdaa,0xad8b,
	0x9de8,0x8dc9,0x7c26,0x6c07,0x5c64,0x4c45,0x3ca2,0x2c83,0x1ce0,0x0cc1,
	0xef1f,0xff3e,0xcf5d,0xdf7c,0xaf9b,0xbfba,0x8fd9,0x9ff8,0x6e17,0x7e36,
	0x4e55,0x5e74,0x2e93,0x3eb2,0x0ed1,0x1ef0
};


/**
 ** This and the following routine (CalcCRC) are straight out of the GEOS-SDK.
 ** These routines are used by pccom to calculate the file transfer checksums.
 ** These routines and the CRC table above have been tested exhaustively so
 ** it would be best to leave these alone.  
 **/

unsigned short incCRC(unsigned short crc, char c) {

	return((crc << 8) ^ crcTable[((crc >> 8) ^ c) & 0xff]);
}


unsigned short calcCRC(char *buf, short size, unsigned short checksum) {

	/*
	 * The CRC is for the data part of the packet only.  The CRC
	 * value is passed low byte first. 
	 */
	for (; size > 0; size--) {
		checksum = incCRC(checksum, *buf++);
	}
	return(checksum);
}


/** Read a single byte from tty.
 **/

unsigned char readByte(int tty) {

	char buf[1];

	read(tty, buf, 1);
	return((unsigned char) buf[0]);
}

/** Read a littleendian 16-bit word from tty.
 **/

unsigned short readWord(int tty) {

	unsigned short ret;

	ret = readByte(tty);
	ret |= (readByte(tty)<<8);
	return(ret);
}

/** Read a littleendian 32-bit word from tty.
 **/

unsigned int readDWord(int tty) {

	unsigned int ret;

	ret = readByte(tty);
	ret |= (readByte(tty)<<8);
	ret |= (readByte(tty)<<16);
	ret |= (readByte(tty)<<24);

	return(ret);
}


/** Write a single byte to tty.
 **/

void writeByte(int tty, unsigned char byte) {

	char buf[1];

	buf[0] = (char) byte;
	write(tty, buf, 1);
}

/** Write a littleendian 16-bit word to tty.
 **/

void writeWord(int tty, unsigned short word) {

	writeByte(tty, word & 0x00ff);
	writeByte(tty, (word & 0xff00)>>8);
}

/** Write a littleendian 32-bit word to tty.
 **/

void writeDWord(int tty, unsigned int dword) {

	writeByte(tty, dword & 0x000000ff);
	writeByte(tty, (dword & 0x0000ff00)>>8);
	writeByte(tty, (dword & 0x00ff0000)>>16);
	writeByte(tty, (dword & 0xff000000)>>24);
}


/** Get a file from the Omnigo.
 **/

void getFile(int tty, char buf[], char name[]) {

	char fname[32];
	unsigned short crc;
	int i, fsize;		/* Strictly this should be an unsigned int, but
						 * this makes detecting size underruns easier (and
						 * will only cause problems when the 2Gb Omnigo
						 * is released! :-)).
						 */
	FILE *f;
	unsigned char b;
	char block[1024];
	short bpos;
	int gotfile = 0;

	/* Turn off echoback. */
	write(tty, "\033EBOFF!", 7);

	write(tty, "\033XF\004", 4);
	write(tty, name, strlen(name)+1);

	readByte(tty);			/* Wait for OG_SYNC */

	while (readByte(tty) == OG_SYNC) {
		readByte(tty);		/* Discard undocumented byte */
		for (i=0; (fname[i]=readByte(tty)) != 0; i++)
			;
		writeByte(tty, OG_SYNC);
		fsize = readDWord(tty);
		printf("Getting file '%s' (%u bytes)\n", fname, fsize);
		gotfile = 1;
		if ((f = fopen(fname, "w")) == NULL) {
			fprintf(stderr, "Couldn't open local file '%s'\n", fname);
			exit(1);
		}
		while (fsize>0) {
			if (readByte(tty) == BLOCK_START) {
				bpos = 0;
				while ((b = readByte(tty)) != BLOCK_END) {
					if (b == BLOCK_QUOTE)
						b = readByte(tty) - BLOCK_QUOTE;
					block[bpos] = b;
					bpos++;
					fsize--;
				}
				crc = readWord(tty);
				if (crc == calcCRC(block, bpos, 0)) {
					writeByte(tty, OG_SYNC);
					putchar('#'); fflush(stdout);
					for (i=0; i<bpos; i++)
						putc(block[i], f);
				}
				else {
					writeByte(tty, OG_NAK);
					putchar('x'); fflush(stdout);
				}
			}
			else
				writeByte(tty, OG_NAK);
		}
		if (readWord(tty) == 0)
			writeByte(tty, OG_ACK);
		else
			writeByte(tty, OG_NAK);
		puts("");
		fclose(f);
	}

	/* Turn echoback back on */
	write(tty, "\033EBON!", 6);
	readReply(tty, buf);

	if (!gotfile)
		puts("Could not get file.");
}


/** Put a file onto the Omnigo.
 **/

void putFile(int tty, char buf[], char name[], char rmt[]) {

	FILE *f;
	struct stat stbuf;
	unsigned int len, i, blksize;
	char block[1024];		/* File buffer */
	int sentok;
	unsigned char reply;

	if ((f = fopen(name, "r")) == NULL) {
		printf("Could not open '%s'.\n", name);
		return;
	}
	stat(name, &stbuf);
	len = stbuf.st_size;

	/* Turn off echoback. */
	write(tty, "\033EBOFF!", 7);

	write(tty, "\033XF\001", 4);
	write(tty, rmt, strlen(rmt)+1);

	if (readByte(tty) != OG_SYNC)
		puts("Could not put file.");
	else {
		printf("Putting file '%s' (%u bytes) as '%s'\n", name, len, rmt);
		writeDWord(tty, len);
		while (len>0) {
			blksize = (len > 1024) ? 1024 : len;
			for (i=0; i<blksize; i++)
				block[i] = fgetc(f);
			do {
				writeByte(tty, BLOCK_START);
				for (i=0; i<blksize; i++)
					switch (block[i]) {
					case BLOCK_START: case BLOCK_END: case BLOCK_QUOTE:
						writeByte(tty, BLOCK_QUOTE);
						writeByte(tty, block[i]+BLOCK_QUOTE);
						break;
					default:
						writeByte(tty, block[i]);
					}
				writeByte(tty, BLOCK_END);
				writeWord(tty, calcCRC(block, blksize, 0));
				reply = readByte(tty);
				if (reply == OG_NAK) {
					sentok = 0;
					putchar('x'); fflush(stdout);
				}
				else if (reply == OG_SYNC) {
					sentok = 1;
					putchar('#'); fflush(stdout);
				}
				else {
					fputs("\nError putting file!\n", stderr);
					exit(1);
				}
			} while (!sentok);

			len -= blksize;
		}
		writeWord(tty, 0);
		readByte(tty);				/* Wait for OG_ACK */
	}
	fclose(f);

	/* Turn echoback back on. */
	write(tty, "\033EBON!", 6);
	readReply(tty, buf);
}

